@@ -15,6 +15,7 @@
 #define _LARGEFILE_SOURCE
 #define _LARGEFILE64_SOURCE
 
+#include "config.h"
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #else
@@ -111,11 +112,11 @@ static void determine_fs_stride(ext2_filsys fs)
 		has_sb = ext2fs_bg_has_super(fs, group);
 		if (group == 0 || has_sb != prev_has_sb)
 			goto next;
-		b_stride = fs->group_desc[group].bg_block_bitmap -
-			fs->group_desc[group-1].bg_block_bitmap -
+		b_stride = ext2fs_block_bitmap_loc(fs, group) -
+			ext2fs_block_bitmap_loc(fs, group - 1) -
 			fs->super->s_blocks_per_group;
-		i_stride = fs->group_desc[group].bg_inode_bitmap -
-			fs->group_desc[group-1].bg_inode_bitmap -
+		i_stride = ext2fs_inode_bitmap_loc(fs, group) -
+			ext2fs_inode_bitmap_loc(fs, group - 1) -
 			fs->super->s_blocks_per_group;
 		if (b_stride != i_stride ||
 		    b_stride < 0)
@@ -158,9 +159,9 @@ int main (int argc, char ** argv)
 	int		force_min_size = 0;
 	int		print_min_size = 0;
 	int		fd, ret;
-	blk_t		new_size = 0;
+	blk64_t		new_size = 0;
 	blk64_t		max_size = 0;
-	blk_t		min_size = 0;
+	blk64_t		min_size = 0;
 	io_manager	io_ptr;
 	char		*new_size_str = 0;
 	int		use_stride = -1;
@@ -180,6 +181,7 @@ int main (int argc, char ** argv)
 	setlocale(LC_CTYPE, "");
 	bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
 	textdomain(NLS_CAT_NAME);
+	set_com_err_gettext(gettext);
 #endif
 
 	add_error_table(&et_ext2_error_table);
@@ -256,11 +258,7 @@ int main (int argc, char ** argv)
 		len = 2 * len;
 	}
 
-#ifdef HAVE_OPEN64
-	fd = open64(device_name, O_RDWR);
-#else
-	fd = open(device_name, O_RDWR);
-#endif
+	fd = ext2fs_open_file(device_name, O_RDWR, 0);
 	if (fd < 0) {
 		com_err("open", errno, _("while opening %s"),
 			device_name);
@@ -352,7 +350,7 @@ int main (int argc, char ** argv)
 				device_name);
 			exit(1);
 		}
-		printf(_("Estimated minimum size of the filesystem: %u\n"),
+		printf(_("Estimated minimum size of the filesystem: %llu\n"),
 		       min_size);
 		exit(0);
 	}
@@ -375,7 +373,7 @@ int main (int argc, char ** argv)
 	 * exceed the partition size.
 	 */
 	retval = ext2fs_get_device_size2(device_name, fs->blocksize,
-					&max_size);
+					 &max_size);
 	if (retval) {
 		com_err(program_name, retval,
 			_("while trying to determine filesystem size"));
@@ -384,8 +382,8 @@ int main (int argc, char ** argv)
 	if (force_min_size)
 		new_size = min_size;
 	else if (new_size_str) {
-		new_size = parse_num_blocks(new_size_str,
-					    fs->super->s_log_block_size);
+		new_size = parse_num_blocks2(new_size_str,
+					     fs->super->s_log_block_size);
 		if (new_size == 0) {
 			com_err(program_name, 0,
 				_("Invalid new size: %s\n"), new_size_str);
@@ -408,7 +406,7 @@ int main (int argc, char ** argv)
 
 	if (!force && new_size < min_size) {
 		com_err(program_name, 0,
-			_("New size smaller than minimum (%u)\n"), min_size);
+			_("New size smaller than minimum (%llu)\n"), min_size);
 		exit(1);
 	}
 	if (use_stride >= 0) {
@@ -440,12 +438,12 @@ int main (int argc, char ** argv)
 	if (!force && (new_size > max_size)) {
 		fprintf(stderr, _("The containing partition (or device)"
 			" is only %llu (%dk) blocks.\nYou requested a new size"
-			" of %u blocks.\n\n"), max_size,
+			" of %llu blocks.\n\n"), max_size,
 			fs->blocksize / 1024, new_size);
 		exit(1);
 	}
-	if (new_size == fs->super->s_blocks_count) {
-		fprintf(stderr, _("The filesystem is already %u blocks "
+	if (new_size == ext2fs_blocks_count(fs->super)) {
+		fprintf(stderr, _("The filesystem is already %llu blocks "
 			"long.  Nothing to do!\n\n"), new_size);
 		exit(0);
 	}
@@ -461,7 +459,7 @@ int main (int argc, char ** argv)
 			exit(1);
 		}
 		printf(_("Resizing the filesystem on "
-			 "%s to %u (%dk) blocks.\n"),
+			 "%s to %llu (%dk) blocks.\n"),
 		       device_name, new_size, fs->blocksize / 1024);
 		retval = resize_fs(fs, &new_size, flags,
 				   ((flags & RESIZE_PERCENT_COMPLETE) ?
@@ -478,7 +476,7 @@ int main (int argc, char ** argv)
 		ext2fs_close(fs);
 		exit(1);
 	}
-	printf(_("The filesystem on %s is now %u blocks long.\n\n"),
+	printf(_("The filesystem on %s is now %llu blocks long.\n\n"),
 	       device_name, new_size);
 
 	if ((st_buf.st_size > new_file_size) &&
